# Copyright 2022 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The `sigstore` Python APIs.

For command-line usage of `sigstore`, refer to the `sigstore`
[README](https://github.com/sigstore/sigstore-python).

Otherwise, here are some quick starting points:

* `sigstore.verify`: verifying of Sigstore signatures,
  including flexible policy control
* `sigstore.sign`: creation of Sigstore signatures
"""

__version__ = "3.5.1"
